'''
A package to manipulate and display some random structures,
including meander systems, planar triangulations, and ribbon tilings


Created on May 8, 2021

@author: vladislavkargin
'''

'''
#I prefer blank __init__.py

from . import mndrpy
from . import pmaps
from . import ribbons
'''