'''
A package to generate, manipulate and display meander systems.
It also handles some incidental objects such as permutations,
Dyck paths, and (non-crossing) pairings
'''

'''
#I prefer blank __init__.py for modules

from .Combinatorics import cycleLengths
from .Combinatorics import cyclesOfPerm
from .Combinatorics import enumPropMeanders
from .Combinatorics import rotationNumber
from .Combinatorics import random231perm

from .DyckPaths import randomDyckPath
from .DyckPaths import plotDyckPath
from .DyckPaths import areaDyckPath
from .DyckPaths import valleys
from .DyckPaths import peaks
from .DyckPaths import pathTo231perm

#from .Meander import Meander
from .Meander import areaMeander
from .Meander import randomMeander
from .Meander import randomComb
from .Meander import randomCrossSemiMeander
from .Meander import rainbowMeander
from .Meander import smallRainbowsMeander
from .Meander import drawAsPolygon

from .Pairing import randomPairing
from .Pairing import randomCrossPairing
from .Pairing import addCrossing
from .Pairing import standardPairing
from .Pairing import rainbowPairing
from .Pairing import rainbows
from .Pairing import smallRainbows
from .Pairing import prng2path

from .Utility import genPairings
from .Utility import allPairings
from .Utility import allMeanders
from .Utility import dot
from .Utility import dotMatrix
from .Utility import randomTree
from .Utility import randomCrossMeander
from .Utility import treeToDyckPath
from .Utility import plotRandomProperMeander
from .Utility import makeMeanderFromCycle
'''


